from mext.dispatch import Adapter

__all__ = [
    'as_xml', 'as_xmlnode',
    'EscapedXML', 'XmlEntity', 'NOTAG',
    'XmlDocument', 'XmlElement',
    'xmlelem',
]

as_xmlnode = Adapter('__xmlnode__')

def escape_text(text):
    return text.replace("&", "&amp;").replace(">", "&gt;").replace("<", "&lt;")

def escape_attribute(text):
    text = escape_text(text)
    text = text.replace('\n', '&#10;').replace('\r', '&#13;').replace('\t', '&#9;')
    if '"' not in text:
        return '"%s"' % text
    elif "'" in text:
        return '"%s"' % text.replace('"', "&quot;")
    else:
        return "'%s'" % text


def as_xml(ob):
    r = []
    _write_xml(ob, r.append)
    return u''.join(r)



def _write_xml(ob, w):
    try:
        writer = _writers[ob.__class__]
    except KeyError:
        t = ob.__class__
        if hasattr(t, '_write_xml'):
            writer = _writers[t] = t._write_xml
        elif hasattr(t, '__xml__'):
            w(ob.__xml__())
            return
        else:
            ob = as_xmlnode(ob)
            writer = _write_xml
            #raise TypeError("No _write_xml implementation for %s" % t)
    writer(ob, w)

def _write_xml_nothing(none, w):
    pass

def _write_xml_asis(s, w):
    w(s)

def _write_xml_iter(iter, w):
    for item in iter:
        _write_xml(item, w)


class EscapedXML(unicode):
    __slots__ = ()

def XmlEntity(code):
    if isinstance(code, int):
        code = '#x%X' % code
    return EscapedXML('&%s;' % code)




_writers = {
    type(None): _write_xml_nothing,
    unicode: (lambda s, w: w(escape_text(s))),
    str: (lambda s, w: w(escape_text(unicode(s)))),
    EscapedXML: _write_xml_asis,
    list: _write_xml_iter,
    tuple: _write_xml_iter,
}


class XmlElement(object):
    __slots__ = ['tag', 'attrs', 'body']
    def __init__(self, __tag, __body=None, **kw):
        self.tag = __tag
        self.body = [__body] if __body is not None else []
        self.attrs = dict((key.rstrip('_').replace('_', '-'), value)
                                    for key, value in kw.iteritems())

    def __iadd__(self, item):
        self.body.append(item)
        return self

    def __repr__(self):
        return (self.__class__.__name__
                + ('(%s)' % ', '.join(   [repr(self.tag)]
                                        + ['%s=%r' % attr for attr in self.attrs.iteritems()]))
                + ('[...]' if self.body else ''))

    def __getitem__(self, body):
        if self.body:
            raise ValueError("%s elem already has a body: %r" % (self.tag, self.body))
        elem = XmlElement(self.tag, body)
        elem.attrs = self.attrs.copy()
        return elem

    _nonempty = False

    def _write_xml(self, w):
        tag = self.tag
        if self.attrs:
            w('<'+tag)
            for key, value in self.attrs.iteritems():
                if value is not None:
                    w(u' %s=%s' % (key, escape_attribute(unicode(value))))
        else:
            w('<'+tag)
        if self.body or self._nonempty:
            w(u'>')
            _write_xml_iter(self.body, w)
            w('</'+tag+'>')
        else:
            w(u'/>')






class XmlElementGen(object):
    __slots__ = ['tag', 'empty']
    elem_cls = XmlElement
    def __init__(self, tag):
        self.tag = tag
        self.empty = as_xml(self.elem_cls(self.tag))

    def _write_xml(self, w):
        w(self.empty)

    def __call__(self, **attrs):
        return self.elem_cls(self.tag, **attrs)

    def __getitem__(self, body):
        return self.elem_cls(self.tag, body)
    __iadd__ = __getitem__


class XmlDocument(object):
    mime_type = 'application/xml'
    _xml_decl_template = u'<?xml version="1.0" encoding="%s" ?>\n'
    _xml_decl_unicode = u'<?xml version="1.0" ?>\n'
    def __init__(self, content, encoding='UTF-8'):
        self.content = as_xml(content)
        self.encoding = encoding

    def __unicode__(self):
        return self._xml_decl_unicode + self.content

    def __str__(self):
        return self.encode(self.encoding)

    def __mimedata__(self):
        from mext.mime import MimeTypedData
        return MimeTypedData(str(self), self.mime_type, charset=self.encoding)

    def __webob__(self):
        from webob import Response
        return Response(str(self), content_type=self.mime_type, charset=self.encoding)

    def encode(self, encoding):
        return ((self._xml_decl_template % encoding) + self.content).encode(encoding)





class _AttrGen(object):
    def __getattr__(self, name):
        return XmlElementGen(name)

xmlelem = _AttrGen()





class _NoTag(list):
    def __iadd__(self, item):
        self.append(item)
        return self

    def __repr__(self):
        return '_NoTag(%s)' % super(_NoTag, self).__repr__()


class NOTAG(object):
    def __call__(self):
        return _NoTag()

    @staticmethod
    def __getitem__(item):
        return _NoTag([item])
    __iadd__ = __getitem__

_writers[NOTAG] = _write_xml_nothing
_writers[_NoTag] = _write_xml_iter

NOTAG = NOTAG()
